#include "stdint.h"
#include "unixtypes.h"
#include "stralloc.h"
#include "error.h"
#include "open.h"
#include "fifo.h"
#include "unix.h"

#include "install.h"
#include "install_log.h"

void _exit(int);

static const char FATAL[] = "install: fatal: ";

static int fdsourcedir = -1;

static void puterr(const char *s)
{
  while (*s) {
    if (write(2, s, 1) < 0) _exit(111);
    s++;
  }
}

static void die_sys(const char *x1, const char *x2, const char *x3, const char *x4)
{
  const char *sys_error_str;
  sys_error_str = error_str(errno);

  puterr(FATAL);

  if (x1) puterr(x1);
  if (x2) puterr(x2);
  if (x3) puterr(x3);
  if (x4) puterr(x4);

  puterr(": ");
  puterr(sys_error_str);
  puterr("\n");
  _exit(111);
}

static void die1sys(const char *x1)
{
  die_sys(x1, (const char *)0, (const char *)0, (const char *)0);
}

static void die2sys(const char *x1, const char *x2)
{
  die_sys(x1, x2, (const char *)0, (const char *)0);
}

static void die4sys(const char *x1, const char *x2, const char *x3, const char *x4)
{
  die_sys(x1, x2, x3, x4);
}

void install_init(void)
{
  fdsourcedir = open_read(".");
  if (fdsourcedir < 0) {
    die1sys("unable to open current directory");
  }
  install_log_init(fdsourcedir);
}

void h(const char *home, uid_t uid, gid_t gid, mode_t mode)
{
  if (mkdir(home, 0700) < 0) {
    if (errno != error_exist) {
      die2sys("unable to mkdir ", home);
    }
  }
  if (chown(home, uid, gid) < 0) {
    die2sys("unable to chown ", home);
  }
  if (chmod(home, mode) < 0) {
    die2sys("unable to chmod ", home);
  }
  log_h(home, uid, gid, mode);
}

void d(const char *home, const char *subdir, uid_t uid, gid_t gid, mode_t mode)
{
  if (chdir(home) < 0) {
    die2sys("unable to switch to ", home);
  }
  if (mkdir(subdir, 0700) < 0) {
    if (errno != error_exist) {
      die4sys("unable to mkdir ", home, "/", subdir);
    }
  }
  if (chown(subdir, uid, gid) < 0) {
    die4sys("unable to chown ", home, "/", subdir);
  }
  if (chmod(subdir, mode) < 0) {
    die4sys("unable to chmod ", home, "/", subdir);
  }
  log_d(home, subdir, uid, gid, mode);
}

void p(const char *home, const char *fifo, uid_t uid, gid_t gid, mode_t mode)
{
  if (chdir(home) < 0) {
    die2sys("unable to switch to ", home);
  }
  if (fifo_make(fifo, 0700) < 0) {
    if (errno != error_exist) {
      die4sys("unable to mkfifo ", home, "/", fifo);
    }
  }
  if (chown(fifo, uid, gid) < 0) {
    die4sys("unable to chown ", home, "/", fifo);
  }
  if (chmod(fifo, mode) < 0) {
    die4sys("unable to chmod ", home, "/", fifo);
  }
  log_p(home, fifo, uid, gid, mode);
}

#define BUFFER_SIZE 512
static char buffer[BUFFER_SIZE];

void c(const char *home, const char *subdir, const char *file, uid_t uid, gid_t gid, mode_t mode)
{
  int fdin;
  int fdout;
  ssize_t rlen;
  ssize_t wlen;

  if (fchdir(fdsourcedir) < 0) {
    die1sys("unable to switch back to source directory");
  }

  fdin = open_read(file);
  if (fdin < 0) {
    die2sys("unable to read ", file);
  }

  if (chdir(home) < 0) {
    die2sys("unable to switch to ", home);
  }
  if (chdir(subdir) < 0) {
    die4sys("unable to switch to ", home, "/", subdir);
  }
  if (unlink(file) < 0) {
    if (errno != error_noent) {
      die2sys("unable to remove ", file);
    }
  }
  fdout = open_trunc(file);
  if (fdout < 0) {
    die4sys("unable to write .../", subdir, "/", file);
  }

  rlen = BUFFER_SIZE;
  for (;;) {
    rlen = read(fdin, buffer, (size_t)rlen);
    if (rlen < 0) die2sys("unable to read ", file);
    if (rlen == 0) break;
    wlen = write(fdout, buffer, (size_t)rlen);
    if (wlen != rlen) die4sys("unable to write .../", subdir, "/", file);
  }

  close(fdin);
  if (fsync(fdout) < 0) {
    die4sys("unable to write .../", subdir, "/", file);
  }
  if (close(fdout) < 0) { /* NFS silliness */
    die4sys("unable to write .../", subdir, "/", file);
  }

  if (chown(file, uid, gid) < 0) {
    die4sys("unable to chown .../", subdir, "/", file);
  }
  if (chmod(file, mode) < 0) {
    die4sys("unable to chmod .../", subdir, "/", file);
  }
  log_c(home, subdir, file, uid, gid, mode);
}

void z(const char *home, const char *subdir, const char *file, unsigned int len, uid_t uid, gid_t gid, mode_t mode)
{
  unsigned int i;
  int fdout;

  if (chdir(home) < 0) {
    die2sys("unable to switch to ", home);
  }
  if (chdir(subdir) < 0) {
    die4sys("unable to switch to ", home, "/", subdir);
  }
  if (unlink(file) < 0) {
    if (errno != error_noent) {
      die2sys("unable to remove ", file);
    }
  }
  fdout = open_trunc(file);
  if (fdout < 0) {
    die4sys("unable to write .../", subdir, "/", file);
  }

  for (i = 0; i < len; i++) {
    if (write(fdout, "", 1) < 0) {
      die4sys("unable to write .../", subdir, "/", file);
    }
  }

  if (fsync(fdout) < 0) {
    die4sys("unable to write .../", subdir, "/", file);
  }
  if (close(fdout) < 0) { /* NFS silliness */
    die4sys("unable to write .../", subdir, "/", file);
  }

  if (chown(file, uid, gid) < 0) {
    die4sys("unable to chown .../", subdir, "/", file);
  }
  if (chmod(file, mode) < 0) {
    die4sys("unable to chmod .../", subdir, "/", file);
  }
  log_z(home, subdir, file, len, uid, gid, mode);
}
