#include <sys/types.h>
#include <sys/stat.h>

#include "stdint.h"
/* #include "unixtypes.h" */
#include "error.h"
#include "unix.h"
#include "macro_unused.h"

#include "install.h"

void _exit(int);

extern struct strerr strerr_sys;

static const char FATAL[] = "instcheck: fatal: ";
static const char WARNING[] = "instcheck: warning: ";

static void puterr(const char *s)
{
  while (*s) {
    if (write(2, s, 1) < 0) _exit(111);
    s++;
  }
}

static void warn5x(const char *x1, const char *x2, const char *x3, const char *x4, const char *x5)
{
  puterr(WARNING);
  puterr(x1);
  puterr(x2);
  puterr(x3);
  puterr(x4);
  puterr(x5);
  puterr("\n");
}

static void warn2sys(const char *x1, const char *x2)
{
  const char *sys_error_str;
  sys_error_str = error_str(errno);
  puterr(WARNING);
  puterr(x1);
  puterr(x2);
  puterr(": ");
  puterr(sys_error_str);
  puterr("\n");
}

static void die_sys(const char *x1, const char *x2, const char *x3, const char *x4)
{
  const char *sys_error_str;
  sys_error_str = error_str(errno);

  puterr(FATAL);

  if (x1) puterr(x1);
  if (x2) puterr(x2);
  if (x3) puterr(x3);
  if (x4) puterr(x4);

  puterr(": ");
  puterr(sys_error_str);
  puterr("\n");
  _exit(111);
}

static void die2sys(const char *x1, const char *x2)
{
  die_sys(x1, x2, (const char *)0, (const char *)0);
}

static void die4sys(const char *x1, const char *x2, const char *x3, const char *x4)
{
  die_sys(x1, x2, x3, x4);
}

static void perm(const char *prefix1, const char *prefix2, const char *prefix3, const char *file, unsigned int type, uid_t uid, gid_t gid, mode_t mode)
{
  struct stat st;

  if (stat(file, &st) < 0) {
    if (errno == error_noent) {
      warn5x(prefix1, prefix2, prefix3, file, " does not exist");
    }
    else {
      warn2sys("unable to stat .../", file);
    }
    return;
  }

  if ((uid != (uid_t)-1) && (st.st_uid != uid)) {
    warn5x(prefix1, prefix2, prefix3, file, " has wrong owner");
  }
  if ((gid != (gid_t)-1) && (st.st_gid != gid)) {
    warn5x(prefix1, prefix2, prefix3, file, " has wrong group");
  }
  if ((st.st_mode & 07777) != mode) {
    warn5x(prefix1, prefix2, prefix3, file, " has wrong permissions");
  }
  if ((st.st_mode & S_IFMT) != type) {
    warn5x(prefix1, prefix2, prefix3, file, " has wrong type");
  }
}

void h(const char *home, uid_t uid, gid_t gid, mode_t mode)
{
  perm("", "", "", home, S_IFDIR, uid, gid, mode);
}

void d(const char *home, const char *subdir, uid_t uid, gid_t gid, mode_t mode)
{
  if (chdir(home) < 0) {
    die2sys("unable to switch to ", home);
  }
  perm("", home, "/", subdir, S_IFDIR, uid, gid, mode);
}

void p(const char *home, const char *fifo, uid_t uid, gid_t gid, mode_t mode)
{
  if (chdir(home) < 0) {
    die2sys("unable to switch to ", home);
  }
  perm("", home, "/", fifo, S_IFIFO, uid, gid, mode);
}

void c(const char *home, const char *subdir, const char *file, uid_t uid, gid_t gid, mode_t mode)
{
  if (chdir(home) < 0) {
    die2sys("unable to switch to ", home);
  }
  if (chdir(subdir) < 0) {
    die4sys("unable to switch to ", home, "/", subdir);
  }
  perm(".../", subdir, "/", file, S_IFREG, uid, gid, mode);
}

void z(const char *home, const char *subdir, const char *file, unsigned int len __UNUSED__, uid_t uid, gid_t gid, mode_t mode)
{
  if (chdir(home) < 0) {
    die2sys("unable to switch to ", home);
  }
  if (chdir(subdir) < 0) {
    die4sys("unable to switch to ", home, "/", subdir);
  }
  perm(".../", subdir, "/", file, S_IFREG, uid, gid, mode);
}
