#ifndef NEMO_IP6_H
#define NEMO_IP6_H

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
  byte_t d[16];
  uint32_t sid;
} ip6_address;

#define IP6_ADDRESS {{'\0','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0'},0}

/*
 ip6 address syntax: (h = hex digit), no leading '0' required
   1. hhhh:hhhh:hhhh:hhhh:hhhh:hhhh:hhhh:hhhh
   2. any number of consecative 0000 may be abbreviated as "::", but only once
 flat ip6 address syntax:
   hhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh
*/

unsigned int	ip6_fmt(const ip6_address *ip, void *out);
unsigned int	ip6_fmt_flat(const ip6_address *ip, void *out);
unsigned int	ip6_fmt_uncompressed(const ip6_address *ip, void *out);

unsigned int	ip6_scan(ip6_address *ip, const char *s);
unsigned int	ip6_scan_flat(ip6_address *ip, const char *s);
unsigned int	ip6_scanbracket(ip6_address *ip, const char *s);
unsigned int	ip6_port_scan(ip6_address *ip, unsigned int *port, const char *s);
unsigned int	ip6_mask_scan(ip6_address *ip, unsigned int *mask_depth, const char *s);
unsigned int	ip6_mask_scanbracket(ip6_address *ip, unsigned int *mask_depth, const char *s);

int		ip6_diff(const ip6_address *ip1, const ip6_address *ip2);
unsigned int	ip6_equal(const ip6_address *ip1, const ip6_address *ip2);
unsigned int	ip6_equal_mask(const ip6_address *ip1, unsigned int mask_depth, const ip6_address *ip2);
unsigned int	ip6_is_v4_mapped(const ip6_address *ip);

void		ip6_zero(ip6_address *ip);
void		ip6_copy(ip6_address *out, const ip6_address *in);

void		ip6_pack(const ip6_address *ip, void *d);
void		ip6_unpack(ip6_address *ip, const void *d);

void		ip6_reverse(ip6_address *ip);

void		ip6_shift_left(ip6_address *ip);
void		ip6_shift_right(ip6_address *ip);
void		ip6_mask_0(ip6_address *ip, unsigned int mask_depth);

void		ip6_copyin_ip4(ip6_address *out, const ip4_address *in);  /* last 4 bytes only */
void		ip6_copyout_ip4(const ip6_address *in, ip4_address *out);  /* last 4 bytes only */

/*
  enough space to hold:
    fully expanded IPv6 notation,
    plus '%'
    plus uint32_t scope id
    plus \0
*/
#define IP6_FMT 51

extern const ip6_address *null_ip6;  /* V6any */
extern const ip6_address *localhost_ip6;  /* V6loopback */
extern const ip6_address *localnet_ip6;
extern const ip6_address *mcastprefix_ip6;
extern const ip6_address *allnodes_ip6;
extern const ip6_address *allrouters_ip6;
extern const ip6_address *allhosts_ip6;
extern const ip6_address *ip6_v4_mapped_prefix;

typedef struct {
  ip6_address *va;
  unsigned int len;
  unsigned int a;
} ip6_vector;

#define IP6_VECTOR {0,0,0}

static inline void ip6_vector_init(register ip6_vector *v)
{
  v->va = 0;
  v->a = v->len = 0;
}

/*  return: 1 = success, 0 = fail  */
unsigned int ip6_vector_realloc(ip6_vector *v, unsigned int n);
unsigned int ip6_vector_alloc(ip6_vector *v, unsigned int n);

static inline unsigned int ip6_vector_ready(register ip6_vector *v, register unsigned int n)
{
  if (v->va) {
    if (n <= v->a) return 1;
    return ip6_vector_realloc(v, n);
  }
  return ip6_vector_alloc(v, n);
}

static inline unsigned int ip6_vector_readyplus(register ip6_vector *v, register unsigned int n)
{
  if (v->va) {
    n += v->len;
    if (n <= v->a) return 1;
    return ip6_vector_realloc(v, n);
  }
  return ip6_vector_alloc(v, n);
}

unsigned int	ip6_vector_cat(ip6_vector *to, const ip6_vector *from);
unsigned int	ip6_vector_copy(ip6_vector *to, const ip6_vector *from);

unsigned int	ip6_vector_append(ip6_vector *v, const ip6_address *i);
unsigned int	ip6_vector_erase(ip6_vector *v);

void		ip6_vector_remove(ip6_vector *v, unsigned int pos, unsigned int len);
void		ip6_vector_swap(ip6_vector *v, unsigned int i, unsigned int j);
void		ip6_vector_sort(ip6_vector *v);
unsigned int	ip6_vector_find(const ip6_vector *v, const ip6_address *ip);

void		ip6_vector_free(ip6_vector *v);

/*
  return length, safely
*/
static inline unsigned int ip6_vector_len(register const ip6_vector *v)
{
  if (v->va) return v->len;
  return 0;
}

#ifdef __cplusplus
}
#endif

#endif
