#include "stdint.h"
#include "unixtypes.h"
#include "unix.h"
#include "stralloc.h"
#include "macro_unused.h"

void _exit(int);

static unsigned int errors = 0;

static void outs(const char *s)
{
  while (*s) {
    if (write(1, s, 1) < 0) _exit(111);
    s++;
  }
}

static void die_nomem(void)
{
  outs("fatal: out of memory\n");
  _exit(1);
}

static void err_compare(const char *in1, const char *in2, int expected, int result)
{
  errors++;
  outs("in1='");
  outs(in1);
  outs("'\n");
  outs("in2='");
  outs(in2);
  outs("'\n");
  outs("expected='");
  if (!expected) {
    outs("0");
  }
  else if (expected < 0) {
    outs("-1");
  }
  else {
    outs("+1");
  }
  outs("'\n");
  outs("result='");
  if (!result) {
    outs("0");
  }
  else if (result < 0) {
    outs("-1");
  }
  else {
    outs("+1");
  }
  outs("'\n");
}

static void analyse_result(const char *in1, const char *in2, int expected, int result)
{
  if (expected < 0) {
    if (result >= 0) {
      err_compare(in1, in2, expected, result);
    }
    return;
  }
  if (!expected) {
    if (result) {
      err_compare(in1, in2, expected, result);
    }
    return;
  }
  /* expected > 0 */
  if (result <= 0) {
    err_compare(in1, in2, expected, result);
  }
}

static stralloc d1 = STRALLOC;
static stralloc d2 = STRALLOC;

static void test_case_diff(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  if (!stralloc_copys(&d2, in2)) die_nomem();
  result = stralloc_case_diff(&d1, &d2);
  analyse_result(in1, in2, expected, result);
}

static void test_case_diffs(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  result = stralloc_case_diffs(&d1, in2);
  analyse_result(in1, in2, expected, result);
}

static void test_case_end(const char *in, const char *end, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in)) die_nomem();
  if (!stralloc_copys(&d2, end)) die_nomem();
  result = (int)stralloc_case_end(&d1, &d2);
  if (result != expected) err_compare(in, end, expected, result);
}

static void test_case_equal(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  if (!stralloc_copys(&d2, in2)) die_nomem();
  result = (int)stralloc_case_equal(&d1, &d2);
  analyse_result(in1, in2, expected, result);
}

static void test_case_equals(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  result = (int)stralloc_case_equals(&d1, in2);
  analyse_result(in1, in2, expected, result);
}

static void test_case_start(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  if (!stralloc_copys(&d2, in2)) die_nomem();
  result = (int)stralloc_case_start(&d1, &d2);
  analyse_result(in1, in2, expected, result);
}

static void test_case_starts(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  result = (int)stralloc_case_starts(&d1, in2);
  analyse_result(in1, in2, expected, result);
}

static void test_diff(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  if (!stralloc_copys(&d2, in2)) die_nomem();
  result = stralloc_diff(&d1, &d2);
  analyse_result(in1, in2, expected, result);
}

static void test_diffb(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  if (!stralloc_copys(&d2, in2)) die_nomem();
  result = stralloc_diffb(&d1, d2.s, d2.len);
  analyse_result(in1, in2, expected, result);
}

static void test_diffs(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  result = stralloc_diffs(&d1, in2);
  analyse_result(in1, in2, expected, result);
}

static void test_end(const char *in, const char *end, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in)) die_nomem();
  if (!stralloc_copys(&d2, end)) die_nomem();
  result = (int)stralloc_end(&d1, &d2);
  if (result != expected) err_compare(in, end, expected, result);
}

static void test_endb(const char *in, const char *end, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in)) die_nomem();
  if (!stralloc_copys(&d2, end)) die_nomem();
  result = (int)stralloc_endb(&d1, d2.s, d2.len);
  if (result != expected) err_compare(in, end, expected, result);
}

static void test_ends(const char *in, const char *end, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in)) die_nomem();
  result = (int)stralloc_ends(&d1, end);
  if (result != expected) err_compare(in, end, expected, result);
}

static void test_equal(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  if (!stralloc_copys(&d2, in2)) die_nomem();
  result = (int)stralloc_equal(&d1, &d2);
  analyse_result(in1, in2, expected, result);
}

static void test_equalb(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  if (!stralloc_copys(&d2, in2)) die_nomem();
  result = (int)stralloc_equalb(&d1, d2.s, d2.len);
  analyse_result(in1, in2, expected, result);
}

static void test_equals(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  result = (int)stralloc_equals(&d1, in2);
  analyse_result(in1, in2, expected, result);
}

static void test_start(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  if (!stralloc_copys(&d2, in2)) die_nomem();
  result = (int)stralloc_start(&d1, &d2);
  analyse_result(in1, in2, expected, result);
}

static void test_startb(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  if (!stralloc_copys(&d2, in2)) die_nomem();
  result = (int)stralloc_startb(&d1, d2.s, d2.len);
  analyse_result(in1, in2, expected, result);
}

static void test_starts(const char *in1, const char *in2, int expected)
{
  int result;
  if (!stralloc_copys(&d1, in1)) die_nomem();
  result = (int)stralloc_starts(&d1, in2);
  analyse_result(in1, in2, expected, result);
}

int main(int argc __UNUSED__, char **argv __UNUSED__)
{
  test_case_diff("fred", "fred", 0);
  test_case_diff("fred", "fred1", -1);
  test_case_diff("fred1", "fred", 1);
  test_case_diff("Fred", "fred", 0);
  test_case_diff("Fred", "fred1", -1);
  test_case_diff("Fred1", "fred", 1);

  test_case_diffs("fred", "fred", 0);
  test_case_diffs("fred", "fred1", -1);
  test_case_diffs("fred1", "fred", 1);
  test_case_diffs("Fred", "fred", 0);
  test_case_diffs("Fred", "fred1", -1);
  test_case_diffs("Fred1", "fred", 1);

  test_case_end("fred", "ed", 1);
  test_case_end("fred", "ed1", 0);
  test_case_end("fred", "ED", 1);
  test_case_end("fred", "ED1", 0);
  test_case_end("fred", "FRED1", 0);

  test_case_equal("fred", "fred", 1);
  test_case_equal("fred", "fred1", 0);
  test_case_equal("fred", "FRED", 1);
  test_case_equal("fred", "FRED1", 0);

  test_case_equals("fred", "fred", 1);
  test_case_equals("fred", "fred1", 0);
  test_case_equals("fred", "FRED", 1);
  test_case_equals("fred", "FRED1", 0);

  test_case_start("fred", "fr", 1);
  test_case_start("fred", "fr1", 0);
  test_case_start("fred", "FR", 1);
  test_case_start("fred", "FR1", 0);

  test_case_starts("fred", "fr", 1);
  test_case_starts("fred", "fred1", 0);
  test_case_starts("fred", "FR", 1);
  test_case_starts("fred", "FRED1", 0);

  test_diff("fred", "fred", 0);
  test_diff("fred", "fred1", -1);
  test_diff("fred1", "fred", 1);
  test_diff("Fred", "fred", -1);
  test_diff("Fred", "fred1", -1);
  test_diff("Fred1", "fred", -1);

  test_diffb("fred", "fred", 0);
  test_diffb("fred", "fred1", -1);
  test_diffb("fred1", "fred", 1);
  test_diffb("Fred", "fred", -1);
  test_diffb("Fred", "fred1", -1);
  test_diffb("Fred1", "fred", -1);

  test_diffs("fred", "fred", 0);
  test_diffs("fred", "fred1", -1);
  test_diffs("fred1", "fred", 1);
  test_diffs("Fred", "fred", -1);
  test_diffs("Fred", "fred1", -1);
  test_diffs("Fred1", "fred", -1);

  test_end("fred", "ed", 1);
  test_end("fred", "ed1", 0);
  test_end("fred", "ED", 0);
  test_end("fred", "ED1", 0);
  test_end("fred", "FRED1", 0);

  test_endb("fred", "ed", 1);
  test_endb("fred", "ed1", 0);
  test_endb("fred", "ED", 0);
  test_endb("fred", "ED1", 0);
  test_endb("fred", "FRED1", 0);

  test_ends("fred", "ed", 1);
  test_ends("fred", "ed1", 0);
  test_ends("fred", "ED", 0);
  test_ends("fred", "ED1", 0);
  test_ends("fred", "FRED1", 0);

  test_equal("fred", "fred", 1);
  test_equal("fred", "fred1", 0);
  test_equal("fred", "FRED", 0);
  test_equal("fred", "FRED1", 0);

  test_equalb("fred", "fred", 1);
  test_equalb("fred", "fred1", 0);
  test_equalb("fred", "FRED", 0);
  test_equalb("fred", "FRED1", 0);

  test_equals("fred", "fred", 1);
  test_equals("fred", "fred1", 0);
  test_equals("fred", "FRED", 0);
  test_equals("fred", "FRED1", 0);

  test_start("fred", "fr", 1);
  test_start("fred", "fr1", 0);
  test_start("fred", "FR", 0);
  test_start("fred", "FR1", 0);

  test_startb("fred", "fr", 1);
  test_startb("fred", "fr1", 0);
  test_startb("fred", "FR", 0);
  test_startb("fred", "FR1", 0);

  test_starts("fred", "fr", 1);
  test_starts("fred", "fred1", 0);
  test_starts("fred", "FR", 0);
  test_starts("fred", "FRED1", 0);

  if (errors) {
    outs("test_stralloc_cmp failed tests!\n");
    _exit(2);
  }
  _exit(0);
}
